//	CD Strip by Chris De Salvo, 1994
//	Based on an the framework of Debugger Strip by Glenn R. Howes

//	------------------------------	Includes

#include <A4Stuff.h>
#include <Icons.h>
#include <Folders.h>
#include <ControlStrip.h>

//	------------------------------	Private Definitions

#define kBaseResID			256
#define kMenuID				kBaseResID
#define kIconSuite			kBaseResID + 1
#define kCDInString			1
#define kCDOutString		2

//	Feature flags
#define kWantMouseClicks	0x0001
#define kDontAutoTrack		0x0002
#define kHasCustomHelp		0x0004
#define kKeepModuleLocked	0x0008

//	------------------------------	Private Constants
//	------------------------------	Private Types
//	------------------------------	Private Structs

typedef struct CDIndex
{
	char	numTracks;
	char	minutes;
	char	seconds;
	char	frames;
	short	strIndex;
} CDIndex;

typedef struct MyGlobals
{
	Handle		iconSuite;			//	Our module's icon
	Boolean		driverInited;		//	Have we successfully opened the CD-ROM driver
	short		driverRef;			//	The driver reference number
	MenuHandle	trackMenu;			//	Menu handle for our pop-up menu
	Str255		cdInHelp;			//	Help balloon text when cd is inserted
	Str255		cdOutHelp;			//	Help balloon text when no cd is inserted
	FSSpec		prefsFile;			//	Location of the 'CD Remote Programs' file
} MyGlobals, *MyGlobalPtr, **MyGlobalHandle;

//	------------------------------	Private Variables
//	------------------------------	Private Functions

static void DoCSDraw(MyGlobalHandle myGlobals, Rect *statusRect);
static void DoCSClose(MyGlobalHandle myGlobals);
static long DoCSInit(void);
static void DoCSClick(MyGlobalHandle myGlobals, Rect *statusRect);
static OSErr DoCSHelp(MyGlobalHandle myGlobals, Rect *statusRect);

static Boolean IsCDInserted(short driverRef);
static short GetNumTracks(short driverRef);
static void PlayTrackNum(int trackNum, short driverRef);
static void ClearMenu(MenuHandle theMenu);
static Boolean FindTitleIndex(MyGlobalHandle myGlobals);
static char GetCurrentTrack(MyGlobalHandle myGlobals);

//	------------------------------	Public Functions

pascal long main (long message, long params, Rect *statusRect, GrafPtr statusPort);

//	------------------------------	Public Variables

//	------------------------------	main

pascal long
main (long message, long params, Rect *statusRect, GrafPtr statusPort)
{
long	result = 0L;
long	oldA4;

	oldA4 = SetCurrentA4();

	switch (message)
	{
		case sdevInitModule:
			result = DoCSInit();
			break;
		
		case sdevCloseModule:
			DoCSClose((MyGlobalHandle) params);
			break;
		
		case sdevFeatures:
			result = kWantMouseClicks | kDontAutoTrack | kHasCustomHelp;
			break;
		
		case sdevGetDisplayWidth:
			result = 16L;			//	Size of a small icon width
			break;
		
		case sdevPeriodicTickle:
			//	This might eventually check for audio-play completion so that we can
			//	do looping.
			break;
		
		case sdevDrawStatus:
			DoCSDraw((MyGlobalHandle) params, statusRect);
			break;
		
		case sdevMouseClick:
			DoCSClick((MyGlobalHandle) params, statusRect);
			break;
		
		case sdevSaveSettings:
			break;
		
		case sdevShowBalloonHelp:
			DoCSHelp((MyGlobalHandle) params, statusRect);
			break;
	}

	SetA4(oldA4);

	return (result);
}

//	------------------------------	DoCSClick

static void
DoCSClick(MyGlobalHandle myGlobals, Rect *moduleRect)
{
short	theTrack;
OSErr	theErr;
char	lockState;
short	refNum;
int		numTracks;
int		loop;
Str255	theString;

	lockState = HGetState((Handle) myGlobals);
	HLock((Handle) myGlobals);
	
	//	Try to open the CD-ROM driver.  The reason that I'm having it attempt this
	//	here instead of in the init section is because Control Strip could load before
	//	The CD-ROM driver extension.  Remember, extensions are loaded alphabetically and
	//	if you're running a third-party driver like DriveCD you'd load after Control Strip.
	if (! (*myGlobals)->driverInited)
	{
		theErr = OpenDriver("\p.AppleCD", &refNum);
		if (theErr)
		{
			HSetState((Handle) myGlobals, lockState);
			return;
		}

		(*myGlobals)->driverInited = TRUE;
		(*myGlobals)->driverRef = refNum;
	}

	//	Find out how many tracks are on the CD and prepare the pop-up menu
	numTracks = GetNumTracks((*myGlobals)->driverRef);
	ClearMenu((*myGlobals)->trackMenu);

	//	If no CD is inserted, then display the error message
	if ((! numTracks) || (! IsCDInserted((*myGlobals)->driverRef)))
	{
		//	Set the menu item to our error message and gray it out
		SetItem((*myGlobals)->trackMenu, 1, "\pNo Audio CD Available");
		DisableItem((*myGlobals)->trackMenu, 1);
	}
	//	If track titles have been entered in CD Remote then use those entries and mark current track.
	//	Otherwise, just use the numbers of the tracks
	else if (FindTitleIndex(myGlobals))
		SetItemMark((*myGlobals)->trackMenu, GetCurrentTrack(myGlobals), '');
	else
	{
		//	I'm just using the track number as menu entries for now until I figure
		//	out how to index into the 'CD Remote Programs' file for the track names.
		NumToString(1L, theString);
		SetItem((*myGlobals)->trackMenu, 1, theString);

		for (loop = 2; loop < numTracks + 1; loop++)
		{
			NumToString((long) loop, theString);
			InsMenuItem((*myGlobals)->trackMenu, theString, loop - 1);
		}
	}

	//	Put up the pop-up menu start tracking
	theTrack = SBTrackPopupMenu(moduleRect, (*myGlobals)->trackMenu);

	//	If there are tracks to play and the user picked one, play it.
	if ((numTracks > 0) && theTrack)
		PlayTrackNum(theTrack & 0xFF, (*myGlobals)->driverRef);

	//	Turn item one back on if it was disabled
	EnableItem((*myGlobals)->trackMenu, 1);
	HSetState((Handle) myGlobals, lockState);
}

//	------------------------------	DoCSInit

static long
DoCSInit()
{
MyGlobalHandle	myH;
Handle			tempHandle = 0L;
OSErr			theErr;
FSSpec			theFSSpec;

	myH = (MyGlobalHandle) NewHandleClear(sizeof (MyGlobals));
	if (! myH)
		return (-1L);

	//	Get a detached handle to the icon suite
	SBGetDetachIconSuite(&tempHandle, kIconSuite, 0x0000FF00);
	(*myH)->iconSuite = tempHandle;

	(*myH)->driverInited = FALSE;

	//	Get a handle to the menu and detach it
	(*myH)->trackMenu = GetMenu(kMenuID);
	if (! (*myH)->trackMenu)
		return (-1L);
		
	DetachResource((Handle) (*myH)->trackMenu);

	//	Get a handle to the balloon help strings, driver name and error message and detach them
	GetIndString((*myH)->cdInHelp, kBaseResID, kCDInString);
	GetIndString((*myH)->cdOutHelp, kBaseResID, kCDOutString);

	//	Find the location of the Preferences Folder so we can look for programs
	theErr = FindFolder(kOnSystemDisk, kPreferencesFolderType, FALSE, &theFSSpec.vRefNum, &theFSSpec.parID);
	if (theErr)
	{
		//	If we couldn't find the folder then fill in the FSSpec with bogus values
		theFSSpec.vRefNum = -1;
		theFSSpec.parID = -1;
		theFSSpec.name[0] = 0;
	}
	else
	{
		//	Complete the FSSpec
		BlockMove("\pCD Remote Programs", &theFSSpec.name, 19);
	}
	
	BlockMove(&theFSSpec, &(*myH)->prefsFile, sizeof (FSSpec));
	
	//	If we made it here, we're successful.  Return our handle.
	return ((long) myH);
}

//	------------------------------	DoCSClose

static void
DoCSClose(MyGlobalHandle myGlobals)
{
	//	We've been asked to close up so free our memory
	//	We're not closing the CD-ROM driver because someone else is probably using it
	if (myGlobals) 
	{
		if ((*myGlobals)->iconSuite)
			DisposeIconSuite ((*myGlobals)->iconSuite, TRUE);

		DisposeHandle((Handle) myGlobals);
	}
}

//	------------------------------	DoCSDraw

static void
DoCSDraw(MyGlobalHandle myGlobals, Rect *statusRect)
{
	//	Just draw our icon in our space
	if ((*myGlobals)->iconSuite)
		PlotIconSuite(statusRect, atAbsoluteCenter, ttNone, (*myGlobals)->iconSuite);
}

//	------------------------------	DoCSDraw

static OSErr
DoCSHelp(MyGlobalHandle myGlobals, Rect *statusRect)
{
	//	If we haven't been able to open the driver don't bother checking the drive
	if (! (*myGlobals)->driverInited)
		return (SBShowHelpString(statusRect, (*myGlobals)->cdOutHelp));
	//	If the driver is open, check the drive for a CD
	else if (IsCDInserted((*myGlobals)->driverRef))
		return (SBShowHelpString(statusRect, (*myGlobals)->cdInHelp));
	//	Yes, this is redundant, but it was quick
	else
		return (SBShowHelpString(statusRect, (*myGlobals)->cdOutHelp));
}

//	------------------------------	IsCDInserted

static Boolean
IsCDInserted(short driverRef)
{
ParamBlockRec	params;
OSErr			theErr;
	
	//	On return from a drive status call the csParam fields have the same info
	//	as the regular disk driver DrvSts structure.  The fourth byte in that
	//	structure is TRUE or FALSE depending on whether or not a disk is in the drive.

	params.cntrlParam.ioCompletion = nil;
	params.cntrlParam.ioVRefNum = 1;
	params.cntrlParam.ioCRefNum = driverRef;
	params.cntrlParam.csCode = 8;					//	Drive Status
	
	theErr = PBStatus((ParmBlkPtr) &params, FALSE);
	if (theErr)
		return (FALSE);
	
	return (params.cntrlParam.csParam[1] & 0xFF);
}

//	------------------------------	GetNumTracks

static short
GetNumTracks(short driverRef)
{
ParamBlockRec	params;
OSErr			theErr;
int				loop;
char			lastTrack;
	
	params.cntrlParam.ioCompletion = nil;
	params.cntrlParam.ioVRefNum = 1;
	params.cntrlParam.ioCRefNum = driverRef;
	params.cntrlParam.csCode = 100;					//	Read TOC
	params.cntrlParam.csParam[0] = 1;				//	Type 1 TOC lookup marker

	//	CD-ROM driver ref says we should zero the rest of the fields out.  Probably don't
	//	but why takes chances?
	for (loop = 1; loop < 11; loop++)
		params.cntrlParam.csParam[loop] = 0;
	
	theErr = PBControl(&params, FALSE);
	if (theErr)
		return (0);

	lastTrack = params.cntrlParam.csParam[0] & 0xFF;

	//	Return the number of tracks in decimal instead of BCD form.
	return ((((lastTrack >> 4) & 0x0F) * 10) + (lastTrack & 0x0F));
}

//	------------------------------	PlayTrackNum

static void
PlayTrackNum(int trackNum, short driverRef)
{
ParamBlockRec	params;
OSErr			theErr;
int				loop;
	
	params.cntrlParam.ioCompletion = nil;
	params.cntrlParam.ioVRefNum = 1;
	params.cntrlParam.ioCRefNum = driverRef;
	params.cntrlParam.csCode = 104;					//	Audio Play
	params.cntrlParam.csParam[0] = 0x0002;			//	Use a BCD Track Number
	
	//	Params 1 and 2 are the 32-bit value (in BCD) of the track to play.
	//	Seems odd to me that they used 2 bytes for this since the highest legal
	//	track number you can have is 99.
	params.cntrlParam.csParam[1] = 0x0000;
	params.cntrlParam.csParam[2] = ((trackNum / 10) << 4) + (trackNum % 10);

	params.cntrlParam.csParam[3] = 0x0000;			//	Indicate that this is a starting address
	params.cntrlParam.csParam[4] = 0x0009;			//	Play in normal stereo mode

	//	Clear the rest of the params
	for (loop = 5; loop < 11; loop++)
		params.cntrlParam.csParam[loop] = 0;
	
	theErr = PBControl(&params, FALSE);
	if (theErr)
		SysBeep(3);
}

//	------------------------------	ClearMenu

static void
ClearMenu(MenuHandle theMenu)
{
short	numItems;

	//	See how many items are currently in the menu
	numItems = CountMItems(theMenu);
	
	//	Remove all but one entry from the menu
	while (--numItems)
		DelMenuItem(theMenu, 1);
	
	//	Clear the item mark if it was there
	SetItemMark(theMenu, 1, 0x00);
}

//	------------------------------	FindTitleIndex

static Boolean
FindTitleIndex(MyGlobalHandle myGlobals)
{
ParamBlockRec	params;
CDIndex			tempIndex;
Handle			theHandle;
int				loop;
short			numEntries;
OSErr			theErr;
short			fileRef;
Ptr				index;
char			minutes, seconds, frames;
short			tracks;
short			strIndex = 0;
Str255			tempString;

	//	Attempt to open the CD Remote Programs file
	fileRef = FSpOpenResFile(&(*myGlobals)->prefsFile, fsRdPerm);
	if (fileRef == -1)
		return (FALSE);
		
	//	Load the Indx resource so we can check for entries
	theHandle = GetResource('IndX', 128);
	if (! theHandle)
	{
		FSClose(fileRef);
		return (FALSE);
	}

	//	Get a pointer to the Index info and skip past the version number
	index = *theHandle;
	index += 2;
	
	//	Find out how many entries there are in the catalog
	numEntries = * ((short *) index);
	if (numEntries == 0)
	{
		FSClose(fileRef);
		return (FALSE);
	}

	//	Skip past the number of entries
	index += 2;

	//	Get the info for the currently mounted audio CD
	params.cntrlParam.ioCompletion = nil;
	params.cntrlParam.ioVRefNum = 1;
	params.cntrlParam.ioCRefNum = (*myGlobals)->driverRef;
	params.cntrlParam.csCode = 100;					//	Read TOC
	params.cntrlParam.csParam[0] = 2;				//	Type 2 TOC lookup marker

	//	Clear unneeded fields
	for (loop = 1; loop < 11; loop++)
		params.cntrlParam.csParam[loop] = 0;
	
	//	Call the CD driver to check the lead-out time of the final session
	theErr = PBControl(&params, FALSE);
	if (theErr)
	{
		ReleaseResource(theHandle);
		FSClose(fileRef);
		return (FALSE);
	}

	//	Extract catalog info
	minutes = (params.cntrlParam.csParam[0] >> 8) & 0xFF;
	seconds = params.cntrlParam.csParam[0] & 0xFF;
	frames = (params.cntrlParam.csParam[1] >> 8) & 0xFF;
	tracks = GetNumTracks((*myGlobals)->driverRef);

	//	Loop through all the entries in the Indx resource and check for matches
	for (loop = 0; loop < numEntries; loop++, index += sizeof (CDIndex))
	{
		BlockMove(index, &tempIndex, sizeof (CDIndex));
		if ((tempIndex.numTracks == tracks)
		&&	(tempIndex.minutes == minutes)
		&&	(tempIndex.seconds == seconds)
		&&	(tempIndex.frames == frames))
			strIndex = tempIndex.strIndex;
	}

	ReleaseResource(theHandle);

	if (! strIndex)
		return (FALSE);

	//	Get name of first song title.  String 1 is the name of the album so we get #2
	GetIndString(tempString, strIndex, 2);
	SetItem((*myGlobals)->trackMenu, 1, tempString);
	EnableItem((*myGlobals)->trackMenu, 1);

	//	Fill in the rest of the entries
	for (loop = 3; loop < tracks + 2; loop++)
	{
		GetIndString(tempString, strIndex, loop);
		InsMenuItem((*myGlobals)->trackMenu, tempString, loop - 2);
		EnableItem((*myGlobals)->trackMenu, loop - 2);
	}
	
	FSClose(fileRef);

	return (TRUE);
}

//	------------------------------	GetCurrentTrack

static char
GetCurrentTrack(MyGlobalHandle myGlobals)
{
ParamBlockRec	params;
OSErr			theErr;
int				loop;
char			track;

	//	Get the info for the current track
	params.cntrlParam.ioCompletion = nil;
	params.cntrlParam.ioVRefNum = 1;
	params.cntrlParam.ioCRefNum = (*myGlobals)->driverRef;
	params.cntrlParam.csCode = 101;					//	Read Q Subcodes for current track

	//	Clear unneeded fields
	for (loop = 0; loop < 11; loop++)
		params.cntrlParam.csParam[loop] = 0;
	
	//	Call the CD driver to check the Q codes for the current track
	theErr = PBControl(&params, FALSE);
	if (theErr)
		return (0xFF);

	track = params.cntrlParam.csParam[0] & 0xFF;
	
	//	Convert to decimal (from BCD) and return
	return (((track >> 4) * 10) + (track & 0x0F));
}